@@ -3,8 +3,24 @@
  *
  * Remote login deaemon with authentication and data encryption of channel
  *
+ * Revision History:
+ *
+ *    12-May-94:  Fixed two malloc bugs in gendeskd.c and deslogind.c.
+ *    14-Jun-94:  
+ *        deslogind.c:	Added -w option for login wrapper feature
+ *                      -w option busted on HP-UX; wait for help.
+ *                      Added RHOSTNAME environment variable.
+ *        utmp.c:	added update of wtmp to writeut
+ *        log.c:	ifdef'd return result from openlog
+ *        socket.c:	errno.h inside #undef posix for ENOTSOCK (BSD/386)
+ *        tty.c:	added control terminal setting for BSD/386
+ *        changed "psignal" to "posignal" to avoid BSD name conflict
+ *        added changes for BSDI BSD/386 
+ *        changed include <ctype.h> before <stdlib.h> for BSD/386
+ *        Thanks to Tom Markson <tom@twilight.com> for BSD/386 changes.
+ *        
  * Still to do:
- *   Add a "RHOSTNAME" and "RHOSTIP" to login environment for users.
+ *   Make the -w option work for HP-UX
  *   Add '.' notation to hostname stuff on socket.c to allow IP addr to be used
  *   figure out a way to run commands and pipelines:
  *      it requires a way to make sure the pipes are flushed before exit
@@ -13,12 +29,16 @@
  *   On hpux, can't reuse port numbers for a short while after exit.  Why?
  *   Check permissions on deslogin.users, and deslogin.log; issue warning.
  *   Encrypt the log files too.
- *   Catch all Coredump signals to prevent hostile user from signalling
+ *   Catch all Coredump signals to prevent hostile user from signaling
  *      us and examining the carcass for keys (deslogind, deslogingw).
  *   How do you prevent a debugger from doing a ptrace(PT_ATTACH)? (cant!)
  *   Compression (tricky because you must not impose delays)
  *   Allow socket to be connected directly to resulting process 
  *       (won't happen because the parent is needed to encipher the data stream)
+ *   Figure out a way to improve protection from local users:
+ *      X11 server DISPLAY variable:
+ *         :0 still doesn't help because that creates a unix-domain socket:
+ *         named /usr/spool/sockets/X11/0 
  *
  *   Deslogin:
  *      Make sure that getPassphrase is not reading from unsecure device:
@@ -43,7 +63,7 @@
  *    Socket is available as stdin and stdout
  *
  * Notes:
- *    I have set psignal so that SIGSTOP will not cause SIGCHLD.  This allows
+ *    I have set posignal so that SIGSTOP will not cause SIGCHLD.  This allows
  *    you to STOP a potential intruder in his tracks by sending SIGSTOP to
  *    the process and then attach a debugger to it and examine what is going
  *    on.
@@ -62,6 +82,7 @@
 			/* setsid setgid setpgid setuid */
 			/* _SC_NGROUPS_MAX */
 #include <limits.h>	/* NGROUPS_MAX */
+#include <ctype.h>	/* must be before stdlib.h for BSDI's BSD 386 */
 #include <stdlib.h>	/* atexit exit free malloc */
 #include <stdio.h>	/* (fopen fclose fgets scanf printf etc.); ctermid */
 #include <string.h>	/* memcmp memcpy memset */
@@ -71,7 +92,6 @@
 #include <termios.h>
 #include <signal.h>
 #include <pwd.h>
-#include <ctype.h>
 #include <time.h>	/* time asctime */
 #include <sys/wait.h>	/* WNOHANG */
 #include "pty.h"
@@ -81,7 +101,7 @@
 #include "socket.h"
 #include "log.h"
 #include "deslogin.h"
-#include "psignal.h"
+#include "posignal.h"
 #include "utmp.h"		/* setlogin */
 #include "group.h"		/* getLoginGroups */
 
@@ -163,9 +183,9 @@ extern int askTty();
 extern unsigned long defaultKey[64];
 
 static char ident[] = 
-" @(#) deslogind.c  version 1.02 09-Apr-94 Copyright 1994 by Dave Barrett(barrett@asgard.cs.Colorado.EDU)\n";
+" @(#) deslogind.c  version 1.10 16-Jun-94 Copyright 1994 by Dave Barrett(barrett@asgard.cs.Colorado.EDU)\n";
 static char RCSid[] = 
-" @(#) deslogind.c  $Revision: 1.2 $ $Date: 94/04/09 14:20:03 $\n";
+" @(#) deslogind.c  $Revision: 1.10 $ $Date: 94/06/14 12:17:58 $\n";
 
 char mainProgName[PROGNAME_SIZE];
 
@@ -181,10 +201,19 @@ struct termios oldmodes;
 char *defaultShell = "/bin/ksh";
 
 char *defaultArgv[] = {
-   (char *) 0,			/* substituted with ang0 variable later */
+   (char *) 0,			/* substituted with arg0 variable later */
+   (char *) 0,			/* if -w, then substitudet with loginArg1 */
+   (char *) 0,			/* if -w, then remote host name */
    (char *) 0
 };
 
+/*
+ * Static data for -w option 
+ */
+char loginArg1[] = "-h";
+char loginProg[] = LOGIN_PROG;
+char loginCwd[]  = "/";
+
 void setProgName() {
    pid_t pid = getpid();				/* can't fail */
    sprintf(mainProgName, "%s[%lu]", argcName, (unsigned long) pid);
@@ -391,8 +420,8 @@ char **updateEnv(ep, key, value)
    return res;
 }
 
-char  **setupEnv(userName, shell, cwd, path, tz)
-   char *userName, *cwd, *shell, *path, *tz;
+char  **setupEnv(userName, shell, cwd, path, tz, rhostName)
+   char *userName, *cwd, *shell, *path, *tz, *rhostName;
 {
    register char **envp = updateEnv((char **) 0, (char *) 0, (char *) 0);
 #ifdef DEFAULT_MAIL
@@ -417,6 +446,12 @@ char  **setupEnv(userName, shell, cwd, path, tz)
    envp = updateEnv(envp, "MAIL", 	envLine);
 #endif
 
+   /*
+    * Add this environment variable to help users adjust their own login
+    * scripts to a given remote enviroment and to detect remote logins
+    * via this program.
+    */
+   envp = updateEnv(envp, "RHOSTNAME", 	rhostName);	/* deslogin feature */
    /* 
     * rlogin propigates TERM, baud rate, changes to window sizes
     */
@@ -526,7 +561,7 @@ int handleSignals() {
    /*
     * To make sure we cleanup utmp with atexit if we're aborted
     */
-   chp = (char *) psignal(SIGINT, sigHandler);
+   chp = (char *) posignal(SIGINT, sigHandler);
    if (chp == (char *) SIG_ERR) {
       log("%s: sigaction SIGINT failed--%s\n", progName, ERRMSG);
       return -1;
@@ -534,7 +569,7 @@ int handleSignals() {
    /*
     * Sent by kill command with no arguments
     */
-   chp = (char *) psignal(SIGTERM, sigHandler); 
+   chp = (char *) posignal(SIGTERM, sigHandler); 
    if (chp == (char *) SIG_ERR) {
       log("%s: sigaction SIGTERM failed--%s\n", progName, ERRMSG);
       return -1;
@@ -542,7 +577,7 @@ int handleSignals() {
    /*
     * Controlling terminal hangup, or death of controlling process
     */
-   chp = (char *) psignal(SIGHUP, sigHandler); 
+   chp = (char *) posignal(SIGHUP, sigHandler); 
    if (chp == (char *) SIG_ERR) {
       log("%s: sigaction SIGHUP failed--%s\n", progName, ERRMSG);
       return -1;
@@ -605,7 +640,7 @@ char *serverHandshake(nfd, rhostName, rport, timeout)
 	 progName, nfd, len, PROTOCOL_VERS, len+1, ERRMSG);
       return 0;
    }
-   count = getString(nfd, protover, VERS_SIZE-1, SETUP_TIMEOUT);
+   count = getString(nfd, protover, VERS_SIZE-1, timeout);
    if (count == 0) {
       log("%s: NOPROTO %s:%d\n", progName, rhostName, rport);
       return 0;
@@ -621,6 +656,92 @@ char *serverHandshake(nfd, rhostName, rport, timeout)
    return &protover[0];
 }
 
+/*
+ * Set the supplementary group id's for the specified user with error checking
+ *
+ * Returns: 0 success, -1 if failure (warning issued)
+ */
+int setupGroups(userName) 
+   char *userName;
+{
+   int res = -1, ngroups, groups;	/* number of supplementary group ids */
+   gid_t *gidset;
+
+   /*
+    * Supplementary group ids are inherited from the parent.  They
+    * should be changed if they exist.  Supplementary group id's are
+    * checked against the target gid of chmod(), chown().
+    */
+   ngroups = (int) sysconf(_SC_NGROUPS_MAX);
+   if (ngroups >= 0) {
+      gidset = (gid_t *) malloc((ngroups+1) * sizeof (gid_t));
+      if (gidset == (gid_t *) 0) {
+	 log(
+      "%s: WARNING--couldn't allocate %d entry supplementary group list\n",
+	    progName, ngroups);
+      } else {
+	 groups = getLoginGroups(ngroups, gidset, userName);
+	 if (groups < 0) {
+	    log("%s: WARNING--getLoginGroups(%d, %s) failed--%s\n", 
+	       progName, ngroups, userName, ERRMSG);
+	 } else {
+	    if (groups > ngroups) {
+	       log(
+	       "%s: WARNING--only first %d of %d supplementary groups set\n",
+		  progName, ngroups, res, userName);
+	       groups = ngroups;
+	    }
+	    res = setGroups(groups, gidset);
+	    if (res < 0) {
+	       log("%s: WARNING--setgroups(%d) failed--%s\n", 
+		  progName, groups, ERRMSG);
+	    }
+	 }
+	 free(gidset);
+      }
+   }
+   return res;
+}
+
+
+/*
+ * Set the user's uid, gid, and supplementary gid's  which determine
+ * the capabilities for this process.
+ */
+void setupUser(uid, gid, userName)
+   uid_t uid;
+   gid_t gid;
+   char *userName;
+{
+   int res;
+
+   res = setgid(gid);		/* must setgid before setuid */
+   if (res != 0) {
+      log("%s: setgid to %d failed--%s\n", progName, gid, ERRMSG);
+      exit(1);
+   }
+   /*
+    * Supplementary group ids are inherited from the parent.  They
+    * should be changed if they exist.  Supplementary group id's are
+    * checked against the target gid of chmod(), chown().
+    */
+   res = setupGroups(userName);
+   /* 
+    * If res < 0, setupgroups failed, but that's OK, warning issued.
+    */
+   /* 
+    * We may lose the ability to update the log file with log() after this
+    */
+   res = setuid(uid);
+   if (debug > 1) {
+      log("%s: setuid:  returned %ld\n", progName, (long) res);
+   }
+   if (res != 0) {
+      log("%s: setuid to %d failed--%s\n", progName, uid, ERRMSG);
+      exit(1);
+   }
+}
+
 int main(argc, argv)
    int argc;
    char *argv[];
@@ -631,19 +752,18 @@ int main(argc, argv)
    char *path = USERPATH, *userFile = USER_FILE, *logName = LOG_FILE, buf[2];
    char **cargv = defaultArgv, **cenv = (char **) 0;
    char ruser[USERNAME_SIZE], userName[USERNAME_SIZE], passPhrase[PHRASE_SIZE];
-   char *userPhrase = (char *) 0;
+   char *userPhrase = (char *) 0, *generic = (char *) 0;
    char mname[PTY_NAMESIZE], rhostName[MAX_HOSTNAME];
    unsigned port = 0, rport = 0;
    unsigned connTimeout  = 1000 * INACTIVITY_TIMEOUT;
    unsigned loginTimeout = 1000 * LOGIN_TIMEOUT;	/* login timer */
    unsigned bufSize = 128;			/* connection bufsize */
-   int      masterpty = -1, slavepty = -1, nfd = -1, pty = 1;
+   int      masterpty = -1, slavepty = -1, nfd = -1, pty = 1, stype;
    int 	    pipe0[2], pipe1[2], pipe2[2];
-   int      ngroups, groups, stype;	/* number of supplementary group ids */
    mode_t   newmask = DEFAULT_UMASK, oldmask;
    pid_t    pid, pid_res;
    uid_t    uid;
-   gid_t    gid, *gidset;
+   gid_t    gid;
    time_t   loginTime;
    keyType  key = (keyType) 0, userKey;
 
@@ -659,7 +779,7 @@ int main(argc, argv)
       }
    }
 
-   while ((ch = getopt(argc, argv, "k:Pl:t:i:f:p:dvnc")) != EOF) switch (ch) {
+   while ((ch = getopt(argc, argv, "w:k:Pl:t:i:f:p:dvnc")) != EOF) switch (ch) {
    case 'c': 				   /* user compile'd userFile key */
       userPhrase = (char *) -1;
       break;
@@ -675,23 +795,26 @@ int main(argc, argv)
    case 'l':
       logName = optarg;
       break;
-   case 'i':
+   case 'i':					/* idle timeout */
       count = sscanf(optarg, "%u", &connTimeout);	/* in seconds */
       if (count != 1) goto usage;
       connTimeout *= 1000;
       break;
-   case 't':
+   case 't':					/* passphrase timeout */
       count = sscanf(optarg, "%u", &loginTimeout);	/* in seconds */
       if (count != 1) goto usage;
       loginTimeout *= 1000;
       break;
-   case 'f':
+   case 'f':					/* set userfile */
       userFile = optarg;
       break;
-   case 'p':
+   case 'p':					/* bind to port # */
       count = sscanf(optarg, "%u", &port);
       if (count != 1) goto usage;
       break;
+   case 'w':				/* function as wrapper for login */
+      generic = optarg;	 /* name in userfile with system-wide passphrase */
+      break;
    case 'v':
       verbose++;
       break;
@@ -700,7 +823,7 @@ int main(argc, argv)
       break;
    default:
 usage:	log(
-"usage: %s  [-dn] [-i inactiveSecs ] [-t loginSecs ] [-l logfile ] [-f userfile ] [-k userfileCipherKey ] [-p port ]\n", 
+"usage: %s  [-dn] [-w name ] [-i inactiveSecs ] [-t loginSecs ] [-l logfile ] [-f userfile ] [-k userfileCipherKey ] [-p port ]\n", 
 progName);
       return 1;
    }
@@ -729,6 +852,13 @@ progName);
    if (res < 0) {
       return 1;
    }
+   /*
+    * The remote host name is not trustworthy for several reasons.
+    * We don't use it (or the remote IP address) for authentication.
+    * Relying on the Domain Name System, or the remote IP address
+    * is not something I'm willing to trust.  Here, we use it strictly
+    * for logging, and informational purposes (who, RHOSTNAME, etc).
+    */
    if (getRemoteAddr(0, rhostName, sizeof rhostName, &rport) == 0) {
       nfd = 0; 			/* From inetd.  0 and 1 is our socket. */
       close(2);		/* inetd leaves it open */
@@ -787,7 +917,17 @@ progName);
    /*
     * Minimize the the time the phrase must me visible in memory.
     */
-   res = getUserPhrase(userFile, passPhrase, PHRASE_SIZE, userName, userKey);
+   if (generic) {
+      chp = generic;
+   } else {
+      chp = userName;
+   }
+   res = getUserPhrase(userFile, passPhrase, PHRASE_SIZE, chp, userKey);
+   if (res == 0) {
+      log("%s: NOPHRASE %s@%s:%d->\"%s\"\n", 
+	 progName, ruser, rhostName, rport, chp);
+      return 1;
+   }
    if (res == -1) {
       log("%s: cannot open \"%s\" for user database\n", progName, userFile);
       return 1;
@@ -797,11 +937,6 @@ progName);
 	 progName, userFile);
       return 1;
    }
-   if (res == 0) {
-      log("%s: NOPHRASE %s@%s:%d->\"%s\"\n", 
-	 progName, ruser, rhostName, rport, userName);
-      return 1;
-   }
    key = challenge(nfd, passPhrase, loginTimeout);
    if (key == (keyType) 0) {
       log("%s: BADLOGIN %s@%s:%d->\"%s\"\n", 
@@ -810,11 +945,18 @@ progName);
    }
    memset(passPhrase, '\0', PHRASE_SIZE);
 
-   res = getUserInfo(userName, &shell, &cwd, &uid, &gid);
-   if (res < 0) {
-      log("%s: NOUSER   %s@%s:%d->\"%s\"\n",
-	 progName, ruser, rhostName, rport, userName);
-      return 1;
+   if (!generic) {
+      res = getUserInfo(userName, &shell, &cwd, &uid, &gid);
+      if (res < 0) {
+	 log("%s: NOUSER   %s@%s:%d->\"%s\"\n",
+	    progName, ruser, rhostName, rport, userName);
+	 return 1;
+      }
+   } else {
+      shell = loginProg;
+      cwd   = loginCwd;		/* not used except for log message */
+      uid   = getuid();		/* cannot fail */
+      gid   = getgid();		/* cannot fail */
    }
 
    /*
@@ -835,7 +977,7 @@ progName);
     * We really should log the tty assigned, and pid of the shell as well
     */
    log("%s: Login %s@%s:%d\n",  progName, ruser, rhostName, rport);
-   log("%s: %s uid(%u) gid(%u) %s %s\n", 
+   log("%s: user:%s uid:%u gid:%u cwd:%s shell:%s\n", 
       progName, userName, uid, gid, cwd, shell);
 
    time(&loginTime);
@@ -942,7 +1084,7 @@ progName);
     * dies.  If not, the symptom is that exiting the remote shell will just
     * hang until the INACTIVITY_TIMEOUT (connTimeout) expires.
     */
-   chp = (char *) psignal(SIGCHLD, sigHandler);
+   chp = (char *) posignal(SIGCHLD, sigHandler);
    if (chp == (char *) SIG_ERR) {
       log("%s: sigaction SIGCHLD failed--%s\n", progName, ERRMSG);
       return 1;
@@ -966,12 +1108,13 @@ progName);
 	 log("%s: parent pid %d; child pid %ld\n", 
 	     progName, getpid(), (long) pid);
       }
-      /*
-       * This must be done after we have a pty, and before we to a setuid.
-       * It updates the utmp file so getlogin and who will work corretly.
-       * It should be done only by the parent.
-       */
       if (pty) {
+	 /*
+	  * This must be done after we have a pty, and before we do a setuid.
+	  * It updates the utmp file so getlogin and who will work corretly.
+	  * It should be done only by the parent.
+	  * It's required on hp-ux even for -w (wrapper) option.
+	  */
 	 res = setlogin(sname, userName, rhostName);
 	 if (res < 0) {
 	    log("%s: WARNING--setlogin %s to %s@%s failed--%s\n", 
@@ -1050,18 +1193,30 @@ progName);
 
    /* child */
    setProgName();
-   res = chdir(cwd);
-   if (res != 0) {
-      log("%s: chdir to %s failed \"%s\"--%s\n", progName, cwd, ERRMSG);
-      return 1;
-   }
-   oldmask = umask(newmask);	/* documentation says umask cannot fail! */
 
-   tz = getenv("TZ");		/* may fail by returning NULL; OK though */
-   if (uid == 0) {		/* root user gets special path */
-      path = SYSTEMPATH;
+   if (generic) {
+      cargv[1] = loginArg1;
+      cargv[2] = rhostName;
+   } else {
+      tz = getenv("TZ");	/* may fail by returning NULL; OK though */
+      if (uid == 0) {		/* root user gets special path */
+	 path = SYSTEMPATH;
+      }
+      cenv = setupEnv(userName, shell, cwd, path, tz, rhostName);
+
+      res = chdir(cwd);
+      if (res != 0) {
+	 log("%s: chdir to %s failed \"%s\"--%s\n", progName, cwd, ERRMSG);
+	 return 1;
+      }
+      oldmask = umask(newmask);	/* documentation says umask cannot fail! */
+
+      /*
+       * This code must be placed only in the child because of the atexit call
+       * in the parent for updating utmp.  This changes the uid, and gid now.
+       */
+      setupUser(uid, gid, userName);	/* exits if failure */
    }
-   cenv     = setupEnv(userName, shell, cwd, path, tz);
 
    if ((chp = strrchr(shell, '/')) == (char *) 0) chp = shell-1;
    chp++;
@@ -1074,57 +1229,7 @@ progName);
    arg0[0] = '-';				/* make this a login shell */
    memcpy(&arg0[1], chp, strlen(chp)+1);
    cargv[0] = arg0;
-   /*
-    * This code must be placed only in the child because of the atexit call
-    * in the parent for updating utmp.
-    */
-   res = setgid(gid);		/* must setgid before setuid */
-   if (res != 0) {
-      log("%s: setgid to %d failed--%s\n", progName, gid, ERRMSG);
-      exit(1);
-   }
-   /*
-    * Supplementary group ids are inherited from the parent.  They
-    * should be changed if they exist.  Supplementary group id's are
-    * checked against the target gid of chmod(), chown().
-    */
-   ngroups = (int) sysconf(_SC_NGROUPS_MAX);
-   if (ngroups >= 0) {
-      gidset = (gid_t *) malloc(ngroups+1);
-      if (gidset == (gid_t *) 0) {
-	 log(
-	 "%s: WARNING--couldn't allocate %d entry supplementary group list\n",
-	    progName, ngroups);
-      } else {
-	 groups = getLoginGroups(ngroups, gidset, userName);
-	 if (groups < 0) {
-	    log("%s: WARNING--getLoginGroups(%d, %s) failed--%s\n", 
-	       progName, ngroups, userName, ERRMSG);
-	 } else {
-	    if (groups > ngroups) {
-	       log(
-		  "%s: WARNING--only first %d of %d supplementary groups set\n",
-		  progName, ngroups, res, userName);
-	       groups = ngroups;
-	    }
-	    res = setGroups(groups, gidset);
-	    if (res < 0) {
-	       log("%s: WARNING--setgroups(%d) failed--%s\n", 
-		  progName, groups, ERRMSG);
-	    }
-	 }
-	 free(gidset);
-      }
-   }
 
-   /* 
-    * We may lose the ability to update the log file with log() after this
-    */
-   res = setuid(uid);
-   if (res != 0) {
-      log("%s: setuid to %d failed--%s\n", progName, uid, ERRMSG);
-      exit(1);
-   }
    /*
     * Sect4.3.2:157-177:
     * If the calling process is not a process group leader, the setsid()
@@ -1408,6 +1513,9 @@ progName);
       serr = dupFd(sin, "serr", O_RDWR);
    }
 
+   if (generic) {
+   }
+
    /*
     * Effect of exec:
     *
@@ -1440,7 +1548,12 @@ progName);
     *    profil(2) - profiling disabled
     */
    if (debug) {
-      log("%s: exec \"%s\"\n", progName, shell);
+      if (generic) {
+	 log("%s: exec \"%s\" as \"%s %s %s\"\n", progName, shell, 
+	    cargv[0], cargv[1], cargv[2]);
+      } else {
+	 log("%s: exec \"%s\" as \"%s\"\n", progName, shell, cargv[0]);
+      }
       if (debug > 1) {
 	 log("%s: env:\n", progName);
 	 showEnv("   ", cenv);
